use clap::Parser;
use itertools::Itertools;
use log::{debug, info, warn};
use rand::prelude::IndexedRandom;
use rand::Rng;
use std::error::Error;
use std::fmt::{Debug, Write};
use std::iter::repeat_with;
use std::net::SocketAddr;
use std::ops::RangeInclusive;
use std::sync::OnceLock;
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::net::{TcpListener, TcpStream};
use tokio::{select, task};

/// Mastermind game server
#[derive(Parser, Debug, Clone)]
#[command(version, about, long_about = None)]
struct Args {
    /// Address and port to listen on
    #[arg(short, long, default_value = "[::]:1234")]
    listen: SocketAddr,

    /// Minimum number of slots in each assignment
    #[arg(short, long, default_value_t = 4)]
    slot_min: u16,

    /// Maximum number of slots in each assignment
    #[arg(short = 'S', long, default_value_t = 4)]
    slot_max: u16,

    /// Minimum number of slots in each assignment
    #[arg(short, long, default_value_t = 6)]
    color_min: u16,

    /// Maximum number of slots in each assignment
    #[arg(short = 'C', long, default_value_t = 6)]
    color_max: u16,

    /// Available colors in each assignment
    #[arg(short = 'L', long, default_value = "red,green,blue,yellow,black,white")]
    colors_list: String,
}

#[derive(Debug)]
struct GameConfig {
    slot_count: RangeInclusive<u16>,
    color_count: RangeInclusive<u16>,
    colors: Vec<String>,
}

impl GameConfig {
    fn new(
        slot_count: RangeInclusive<u16>,
        color_count: RangeInclusive<u16>,
        colors: Vec<String>,
    ) -> Self {
        Self {
            slot_count,
            color_count,
            colors,
        }
    }

    fn from_args(
        slot_count: RangeInclusive<u16>,
        color_count: RangeInclusive<u16>,
        colors: String,
    ) -> Self {
        Self::new(
            slot_count,
            color_count,
            colors
                .split(',')
                .map(str::trim)
                .map(str::to_string)
                .collect(),
        )
    }
}

trait LogResult {
    type Value;
    fn log(self) -> Option<Self::Value>;
}

impl<V, E> LogResult for Result<V, E>
where
    E: Debug,
{
    type Value = V;
    fn log(self) -> Option<V> {
        self.map_err(|err| {
            warn!("{err:?}");
        })
        .ok()
    }
}

async fn handle(
    mut stream: TcpStream,
    src_addr: SocketAddr,
    mut exit_rx: tokio::sync::watch::Receiver<bool>,
) {
    info!("({src_addr}) Received a new connection");

    let config = CONFIG.get().expect("Game config is not initialized");

    let selected_colors = config
        .colors
        .choose_multiple(
            &mut rand::rng(),
            rand::rng().random_range(config.color_count.clone()) as usize,
        )
        .map(String::as_str)
        .collect_vec();

    let hidden_sequence = repeat_with(|| *selected_colors.choose(&mut rand::rng()).unwrap())
        .take(rand::rng().random_range(config.slot_count.clone()) as usize)
        .collect_vec();
    debug!("({src_addr}) Generated sequence: {hidden_sequence:?}");
    let hidden_sequence_counts = hidden_sequence.iter().cloned().counts();
    let (read, mut write) = stream.split();
    let mut read = BufReader::new(read);

    let mut buffer = String::new();

    write!(
        buffer,
        "Welcome\n{}\n{}\n",
        hidden_sequence.len(),
        selected_colors.join(",")
    )
    .log();
    write.write_all(buffer.as_bytes()).await.log();
    buffer.clear();

    loop {
        select! {
            _ = exit_rx.changed() => {
                write.write_all("Server exiting...\nGoodbye\n".as_bytes()).await.log();
                break;
            }
            result = read.read_line(&mut buffer) => {
                match result {
                    Err(err) => {
                        buffer.clear();
                        write!(&mut buffer, "Error occured while reading: {err:?}\nGoodbye\n").log();
                        info!("({src_addr}) {buffer}");
                        write.write_all(buffer.as_bytes()).await.log();
                        break;
                    }
                    Ok(_) => {
                        let guess = buffer.split(',')
                            .map(str::trim)
                            .collect_vec();
                        if guess.len() != hidden_sequence.len() {
                            let guess_len = guess.len();
                            drop(guess);
                            buffer.clear();
                            write!(&mut buffer, "Your guess has the wrong number of slots. Correct: {}, given: {}\nGoodbye\n", hidden_sequence.len(), guess_len).log();
                            write.write_all(buffer.as_bytes()).await.log();
                            break;
                        }
                        let correct_positions = guess.iter()
                            .zip(hidden_sequence.iter())
                            .filter(|(lhs, rhs)| lhs == rhs)
                            .count();
                        if correct_positions == hidden_sequence.len() {
                            write.write_all("Correct\nGoodbye\n".as_bytes()).await.log();
                            break;
                        }
                        let incorrect_positions = guess.iter()
                            .cloned()
                            .counts()
                            .into_iter()
                            .map(|(color, count)| {
                                let total_repeats = hidden_sequence_counts.get(color).unwrap_or(&0);
                                count.min(*total_repeats)
                            })
                            .sum::<usize>() - correct_positions;
                        buffer.clear();
                        write!(&mut buffer, "{correct_positions},{incorrect_positions}\n").log();
			debug!("{correct_positions},{incorrect_positions}");
                        write.write_all(buffer.as_bytes()).await.log();
                        buffer.clear();
                    }
                }
            }
        }
    }
}

static CONFIG: OnceLock<GameConfig> = OnceLock::new();

#[tokio::main]
async fn main() -> Result<(), Box<dyn Error>> {
    let args = Args::parse();
    env_logger::init();

    CONFIG
        .set(GameConfig::from_args(
            args.slot_min..=args.slot_max,
            args.color_min..=args.color_max,
            args.colors_list,
        ))
        .expect("Game config already initialized");
    let listener = TcpListener::bind(args.listen).await?;

    info!("Listening on {}", args.listen);

    let mut handles = vec![];
    let (tx, rx) = tokio::sync::watch::channel(false);

    loop {
        select! {
            _ = tokio::signal::ctrl_c() => {
                info!("Exiting...");
                tx.send(true).expect("Failed to send stop signal to handle threads");
                break;
            }
            res = listener.accept() => {
                let (stream, src_addr) = res?;
                handles.push(task::spawn(handle(stream, src_addr, rx.clone())));
            }
        }
    }

    for handle in handles {
        handle.await?;
    }

    Ok(())
}
